open System.IO
open Ast
open Pars
open List
open Printf
open Lexing

let rec readFile (filename:string) : (ParseClause list) =
    printf "parsing %s\n" filename;
    try
    let stream = new StreamReader(filename) in 
    let myProg =
	let lexbuf = Lexing.from_text_reader System.Text.Encoding.ASCII stream in
        try
	    Pars.start Lex.token lexbuf
	with e ->
	    let pos = lexbuf_curr_p lexbuf in 
	    printf "parse error near line %d, character %d; aborting\n" (pos.pos_lnum+1) 
				(pos.pos_cnum - pos.pos_bol);
	    exit 1
    in
    match myProg with
    | Program(Preprocess(r), Parseprogbody(b)) -> 
	let includeFacts = List.flatten (List.map readFile r) in
	(List.append includeFacts b)
    with _ -> printf "ERROR reading file %s; aborting\n" filename; exit 1
;;

let getExtraVar index =
    Var(sprintf "_%d" index)
;;

(* checks for constants on lhs and replaces them by constant function values *)
let checkLhsConstants (l:Literal) (counter:int) =
    match l with | (pred,tuple) ->
    List.fold_right (fun term ((pred,newTuple),newFns,counter) ->
	match term with
	| Var(x) -> ((pred,term::newTuple),newFns,counter)
	| _ -> 	let var = getExtraVar counter in
		let fn = FnEqual(var,Terminal(term)) in
		((pred,var::newTuple),fn::newFns,counter+1)
    ) tuple ((pred,[]),[],counter)
;;

(* checks for variables that occur more than once in positive rhs literals
   and replaces all but the first occurrence by filter conditions *)
let checkRhsDuplicates (l:Literal) (counter:int) =
   match l with | (pred,tuple) ->
   List.fold_right (fun term ((pred,newTuple),newFilters,counter) ->
	match term with
	| Var(x) ->
		if (List.exists (fun e -> (e=term)) newTuple) then begin
		    let var = getExtraVar counter in
		    let filter = FilterEqual(term,var) in
		    ((pred,var::newTuple),filter::newFilters,counter+1)
		end else
		    ((pred,term::newTuple),newFilters,counter)
	| _ -> ((pred,term::newTuple),newFilters,counter)
   ) tuple ((pred,[]),[],counter)
;;

let checkDefinedVars rule =
   match rule with | Rule(lhsLit,posRhs,negRhs,filters,functions) ->
   let filterVarsFromTuple tuple = List.filter (fun x -> match x with | Var(x) -> true | _ -> false) tuple in
   let varsInLit lit = match lit with | (_,tuple) -> Set.of_list (filterVarsFromTuple tuple) in
   let fnLhsVar fn = match fn with | FnEqual(var,_) -> Set.singleton var in
   let fnRhsVars fn =
	let rec fnRhsVarsAux e = match e with
		| Minus(e1,e2) -> Set.union (fnRhsVarsAux e1) (fnRhsVarsAux e2)
		| Plus(e1,e2) -> Set.union (fnRhsVarsAux e1) (fnRhsVarsAux e2)
		| Terminal(Var(x)) -> Set.singleton (Var(x))
		| _ -> Set.empty
	in match fn with | FnEqual(_,e) -> fnRhsVarsAux e in
   let filterVars fl = match fl with
	| LessThan(t1,t2)      -> Set.of_list (filterVarsFromTuple (t1::[t2]))
	| GreaterThan(t1,t2)   -> Set.of_list (filterVarsFromTuple (t1::[t2]))
	| LessThanEqual(t1,t2) -> Set.of_list (filterVarsFromTuple (t1::[t2]))
	| FilterEqual(t1,t2)   -> Set.of_list (filterVarsFromTuple (t1::[t2])) in

   let lhsVars = varsInLit lhsLit in
   let posVars = List.fold_right Set.union (List.map varsInLit posRhs) Set.empty in
   let negVars = List.fold_right Set.union (List.map varsInLit negRhs) Set.empty in
   let fnDef = List.fold_right Set.union (List.map fnLhsVar functions) Set.empty in
   let fnUse = List.fold_right Set.union (List.map fnRhsVars functions) Set.empty in
   let filtVars = List.fold_right Set.union (List.map filterVars filters) Set.empty in

   let testError msg varset =
	if not (Set.is_empty varset) then begin
		let var = Set.min_elt varset in
		(* the following assertion should never fail *)
		let varname = match var with | Var(x) -> x in
		printf "ERROR: Variable %s %s in the following rule:\n  %s\nAborting.\n"
			varname msg (Print.getRuleName rule);
		exit 1
	end in

   let definedVars = Set.union posVars fnDef in
   let usedVars = Set.union (Set.union lhsVars negVars) filtVars in

   testError "redefined" (Set.inter posVars fnDef);
   testError "unbound" (Set.diff fnUse posVars);
   testError "unbound" (Set.diff usedVars definedVars);

   | _ -> (assert(false))	(* this will never happen *)
;;

let parseRule (l:Literal) (plist: ParseLiteral list): (Clause) =
    let (newLhs,extraFns,counter) = checkLhsConstants l 0 in
    let (posList, negList, flList, fnList, _) = 
	List.fold_left (
	    fun (pl1, negl1, fl1, fnl1, ctr) pl ->
		match pl with 
		|PosLiteral(a) ->
		    let (modRhsLit,extraFilters,newCtr) = checkRhsDuplicates a ctr in
		    (modRhsLit::pl1, negl1, List.append extraFilters fl1, fnl1, newCtr)
		|NegLiteral(a) -> (pl1, a::negl1, fl1, fnl1, ctr)
		|FilterLiteral(a) -> (pl1, negl1, a::fl1, fnl1, ctr)
		|FunctionLiteral(a) -> (pl1, negl1, fl1, a::fnl1, ctr)
               ) 
	    ([],[],[],extraFns,counter) plist
    in
	let rule = Rule(newLhs, posList, negList, flList, fnList) in
	checkDefinedVars rule;
	rule
;;

let sortRules biglist =
    List.fold_left (fun (fl,rl,ql) pcl ->
	match pcl with 
	|PFact(a) ->  (Fact(a)::fl, rl, ql)
	|PRule(a,b) -> (fl, (parseRule a b)::rl, ql)
	|PQuery(a) ->
		let (Pred(n), _) = a in
		let rName = sprintf "%s_Query" n (*(List.length ql)*) in
		let rule = Rule((Pred(rName),[]),[a],[],[],[]) in
		(fl, rule::rl, (List.append ql [rule]))
	)
	([],[],[]) biglist
;;

let getRules filename =
	sortRules (readFile filename)
