open Ast
open Microsoft.FSharp.Collections.Set

(* 
	Transforms the parse-time datastrucutures into real datastructures representing
	the datalog program. In particular, the following steps are executed in order.
	
	1) Validates whether semantics are defined for the parse-time datastructure.
	   -- Not all constructs passed through by parsing have semantics. Validate whether
	      the program is well-formed and meets all the constraints like only unary base
	      predicates.
	         --A datalog program is considered valid if and only if
	           --Every query, Edb, new Rule, next Rule , regular Rule is valid.
	         
	         --Query Rule is valid if and only if
	             it is composed of PositiveLiterals or OR or AFTER.
	             (Currently, taken care by the parser)
	         
	         --Edb Rule is valid if and only if
			     it is a Positive Unary Predicate.
				 (Currently, taken care by the parser)
				
			 --Regular Rule is valid if and only if 
			   -- LHS is a single positive predicate (currently taken care by parser)
			   -- all terms that occur in the LHS predicate occur in some predicate on the RHS
			   -- and all negated terms on the RHS 
					-- occur in lhs
					-- occur in some positive predicate on RHS
			   -- and all negation is stratified across rules (NB: deferred after translation)
			    (Look at old code)
			     	
			 --New Rule is valid if and only if all LHS predicates are positive unary predicates 
			   and RHS is a zeroary predicate 
			   (Already done by parser)
			 
			 --Next Rule is valid if and only if 
			   -- LHS contains only unary predicate(s) with exactly one variable
			   -- RHS contains any predicate
			   -- LHS variable must be bound
	
	2) Assigns semantics to well-formed programs. The whole program is transformed
	   to the real-time datastrucutures as defined in "ast.fs".
	   -- The construction of the real-time datastructures involves
	      1) parsing each included text and constructing a list of clause for each file.
	      2) fill the real datastructure
	      
*)

let getTermList (x:System.Object) =
(
	match x with 
	| :? Literal as v ->(
	 	match v with PosLiteral(PosLit(Base(a,(b)))) -> b
		| NegLiteral(NegLit(Base(a,b))) -> b
		| PosUnary(PosUp(BaseUnary(a,b))) -> (b::[])
		| NegUnary(NegUp(BaseUnary(a,b))) -> (b::[])
		| Zeroary(a) -> []
		)
	| :? NegativeLiteral as v -> (
			match v with NegLit(Base(_,b)) -> b
		)
	| :? PositiveLiteral as v -> (
			match v with PosLit(Base(_,b)) -> b
		)
	| :? BaseLiteral as v -> (
			match v with Base(_,b) -> b
		)
	| :? UnaryPredicate as v -> (
			match v with 
			|PUP(PosUp(BaseUnary(_,b))) -> b::[]
			|NUP(NegUp(BaseUnary(_,b))) -> b::[]
		)
	| :? NegativeUnaryPredicate as v -> (
			match v with NegUp(BaseUnary(_,b)) -> b::[]
		)
	| :? PositiveUnaryPredicate as v -> (
			match v with PosUp(BaseUnary(_,b)) -> b::[]
		)
	| :? BaseUnaryPredicate as v -> (
			match v with BaseUnary(_,b) -> b::[]
		)
	| :? ZeroaryPredicate as  v -> []
	| _ -> raise (new System.Exception( "term list not defined") )
)

let getVariablesList (v:'a list) : (Term List) =
(
	let ret = ref [] in
	(
	List.iter (
				fun x -> 
				let y = getTermList(x:>System.Object) in 
				(
					ret := ( List.append (!ret)  y  ) 
				) 
			  ) v ;
	!ret
	)
)

let separatePosNegTerms (v:Literal list) : (Term List * Term List) =
(
	let pos = ref [] in
	let neg = ref [] in
	(
		List.iter (
					fun x -> match x with
					|PosLiteral (a) -> pos := List.append (!pos) ( getTermList( a:>System.Object ) )
					|PosUnary (a) -> pos := List.append (!pos) ( getTermList( a:>System.Object ) )
					|NegUnary (a) -> neg := List.append (!neg) ( getTermList( a:>System.Object ) )
					|NegLiteral (a) -> neg := List.append (!neg) ( getTermList( a:>System.Object ) )
					| _ -> ()
					)	v ;
		(!pos, !neg)
	)
) 

let separatePosNegLiterals (v:Literal List) : (PositiveLiteral List * NegativeLiteral List) =
(
	let pos = ref [] in
	let neg = ref [] in
	(
		List.iter (
					fun x -> match x with
					|PosLiteral (a) -> pos := (a :: (!pos) )
					|PosUnary (PosUp(BaseUnary(a,b))) -> pos := (PosLit(Base(a,(b::[]))) :: (!pos) )
					|NegUnary (NegUp(BaseUnary(a,b))) -> neg := (NegLit(Base(a,(b::[]))) :: (!neg) )
					|NegLiteral (a) -> neg := (a :: (!neg) )
					| _ -> ()
					)	v ;
		(!pos, !neg)
	)

)

let separatePosNegUPs(v:UnaryPredicate List) :  ( PositiveUnaryPredicate List * NegativeUnaryPredicate List ) =
(
	let pos = ref [] in
	let neg = ref [] in
	(
		List.iter (
					fun x -> match x with
					|PUP(a) -> pos := (a :: (!pos) )
					|NUP(a) -> neg := (a :: (!neg) )
					) v;
		(!pos,!neg)
	)
)

let updatePredicateArity predicate arity (currentMap:Tagged.Map<'a,'b,'c> ref) = 
(
	match (!currentMap).TryFind predicate with
	|Some(z) -> ( 
		if(arity = z) then true 
		else (
			raise ( new System.Exception(predicate ^ " is used with different arities " ^ string_of_int(arity) ^" and " ^ string_of_int(z) ) )
		)
	)
	| None -> ( currentMap := (!currentMap).Add(predicate,arity)  ; true)
)

let getPredicateArityListFromLitList (litList:Literal List) =
(
	let ret = ref [] in
	(
		List.iter(
			fun x -> 
			match x with
			|PosLiteral(PosLit(Base(Pred(p),tl))) ->
			ret := (p, tl.Length) :: (!ret)
			|NegLiteral(NegLit(Base(Pred(p),tl))) ->
			ret := (p, tl.Length) :: (!ret)
			|PosUnary(PosUp(BaseUnary(Pred(p),t))) ->
			ret := (p,1)::(!ret)
			|NegUnary(NegUp(BaseUnary(Pred(p),t))) ->
			ret := (p,1)::(!ret)
			|Zeroary(Pred(p)) ->
			ret := (p,0)::(!ret)
		) litList
	);
	!ret
)

let rec getPredicateArityListFromQuery (query:Query) =
(
	let ret = ref [] in
	(
		match query with 
		|BaseQuery(posLit) -> 
			ret := getPredicateArityListFromLitList(PosLiteral(posLit) :: [])
			
		|ORQuery(a,b) |AfterQuery(a,b) ->
			ret := List.concat (getPredicateArityListFromQuery(a):: (getPredicateArityListFromQuery(a)::[]))
		
	);
	!ret
)

let getPredicateArityListFromPUPList (pupList:PositiveUnaryPredicate List)=
(
	let litList = ref [] in
	List.iter(
		fun x -> litList := PosUnary(x) :: (!litList)
	) pupList;
	getPredicateArityListFromLitList(!litList)
)

let getPredicateArityListFromUPList (upList:UnaryPredicate List)=
(
	let litList = ref [] in
	List.iter (
		fun x -> match x with PUP(pup) -> 
			litList := PosUnary(pup)::(!litList)
			|NUP(nup) -> litList := NegUnary(nup)::!litList;
	) upList;
	getPredicateArityListFromLitList(!litList);
)


let getPredicateArityListFromClause (clause:P_Clause) : ( string * int) List =
(
	let ret = ref [] in
	(
		match clause with
		|PEdb(pup) -> 
			let y = getPredicateArityListFromPUPList(pup::[]) in
				ret := List.append (!ret) y
		|PQuery(query) ->
			let y = getPredicateArityListFromQuery(query) in
				ret := List.append (!ret) y
		|PRule(posLit , litList) ->
			let y = getPredicateArityListFromLitList(litList) in(
				match posLit with PosLit(Base(Pred(p),tl)) ->
				ret := List.append (!ret) y;
				ret := List.append (!ret) ((p,tl.Length)::[])
			)
		|PNewRule(pupList , zero) ->
			let y = getPredicateArityListFromPUPList(pupList) in (
				ret := List.append (!ret) y;
				match zero with Pred(p) ->
					ret := (p,0)::(!ret);
			)
		|PNextRule(upList, litList) ->
			let y = getPredicateArityListFromUPList(upList) in
			let z = getPredicateArityListFromLitList(litList) in(
				ret := List.append (!ret) y;
				ret := List.append (!ret) z;
			)
	);
	!ret
)

(* Checks if every predicate P has the same arity across all its uses 
	If succesful, returns a map<predicateName , arity>
*)
let checkPredicateNaming (clauseList:P_Clause List) =
(
	let predicateToArity = ref (Map.MakeTagged (System.StringComparer.Ordinal) ).empty in
	let predicateNamingCheck =
	(
		List.for_all (
			fun x -> (
				let y = getPredicateArityListFromClause(x) in
				List.iter (
					fun z -> match z with (pred,arity) ->(
						let temp = updatePredicateArity pred arity predicateToArity in
						()
					)
				) y;
				true;
			)
		) clauseList;
	) in (
		if(predicateToArity.contents.ContainsKey("Reachable") )
			then raise ( new System.Exception("Reachable cannot be used as predicate name.") );
		predicateToArity.contents.Iterate(fun x y ->
			if(x.StartsWith("Autocontaminated"))
				then raise (new System.Exception("Predicate names cannot start with \"AutoContaminated\""));
			if(x.StartsWith("AutoQuery"))
				then raise (new System.Exception("Predicate names cannot start with \"AutoQuery\""));
		) ;
		if(predicateToArity.contents.ContainsKey("Reachable") )
			then raise ( new System.Exception("Reachable cannot be used as predicate name.") );
		!predicateToArity
	)
)

let validateQuery  (query:Query) = true
let validateEdb   (edb:PositiveUnaryPredicate) = true
let validateRule   (rule:PositiveLiteral*(Literal list)) =
(
	match rule with (a,b) -> 
	(
		let lhsvars = getTermList(a:>System.Object) in
		let rhsvars = getVariablesList ( b) in
		let out = List.for_all ( fun x -> (List.exists (fun y -> (y=x)) rhsvars ) ) lhsvars in
		(
			if out=true then
			(
				let (rhsPosVars, rhsNegVars) = separatePosNegTerms(b) in
				(
				(* ------- THESE TESTS ARE TOO RESTRICTIVE
					let boundTest = List.for_all (fun x->(List.exists (fun y-> y=x) lhsvars )) rhsNegVars in
					let posTest   = List.for_all (fun x->(List.exists (fun y-> y=x) rhsPosVars )) rhsNegVars in
					if boundTest && posTest then true
				*)
					let boundORposTest = List.for_all (fun x -> (List.exists (fun y -> y = x) (List.rev_append lhsvars rhsPosVars))) rhsNegVars in
					if boundORposTest then true
					else (
						printf "ERROR: in regular rule %a\n" output_any (rule);
						failwith "Some negated variable in rhs is either not bound or not found in any positive rhs predicate"
					)
				)
			)
			else (
				printf "ERROR: in regular rule %a\n" output_any (rule);
				failwith "Some lhs variable is not bound " 
			)
		)
	)
)

let validateNewRule (newrule:(PositiveUnaryPredicate List) * ZeroaryPredicate) = true
let validateNextRule (nextrule:(UnaryPredicate List) * (Literal List) ) =
(
	match nextrule with (a,b) ->
	let lhsVars = getVariablesList(a) in
	let rhsVars = getVariablesList(b) in
	(
		let uniqueLhsVariableTest = 
		(
			match lhsVars with  [] -> false
			|head::tail -> List.for_all (fun x -> x=head) tail
		) in
		let boundTest = (
			match lhsVars with [] -> false
			|head::tail -> List.exists (fun x -> x = head) rhsVars
		) in 
		if uniqueLhsVariableTest && boundTest then true
		else
		(
			printf "ERROR: in next rule %a\n" output_any (nextrule);
			failwith "Lhs Variable is not unique or Lhs Variable is not bound."			
		)
	)
)

let validateClause (clause:P_Clause) =
match clause with 
			|PEdb(pup) -> validateEdb (pup)
			|PQuery(query) -> validateQuery(query)
			|PRule(a,b) -> validateRule(a,b)
			|PNewRule(a,b) -> validateNewRule(a,b)
			|PNextRule(a,b) -> validateNextRule(a,b)


let validateAll (clauseList:(P_Clause List) ) = 
List.for_all (validateClause ) clauseList

(* transform a clause of type P_Clause to type Clause *)
let transformPClauseToClause (clause:P_Clause) : (Clause) = 
(
	match clause with 
	|PEdb(a) -> Edb(a)
	|PQuery(a) -> Query(a)
	|PRule(a,b) -> let (c,d) = separatePosNegLiterals(b) in Rule(a,c,d)
	|PNewRule(a,b) -> NewRule(a,b)
	|PNextRule(a,b) -> let (p,q) = separatePosNegUPs(a) in let (r,s) = separatePosNegLiterals(b) in NextRule(p,q,r,s)
)

let transformParseOutToProgram (p:P_Clause List) : (Clause List) =
(
	(
		let clauseList = ref [] in
		List.iter 
		(
			fun x -> let y = transformPClauseToClause x in clauseList := (y::(!clauseList))
		) p ;
		(!clauseList)
	)
)


(* Tests if term list is unary *)
let testTermListUnary(tList : Term List) =
(
let l = tList.Length in
 if l = 1 then true else false
)

let testPredicateNotUnary (map:Tagged.Map<string,int,System.StringComparer>) predicateName = 
(
	match predicateName with Pred(x) ->
	match map.TryFind x with
	|Some(z) -> z<>1
	|None -> true
)

(* Checks if all EDB predicates are unary *)
let checkEDBUnary(edpp, map) = 
( 
  let epl = Set.to_list edpp in
  (* ----- ADDING MORE INFORMATION IN ERROR MESSAGE BELOW
  let test = List.for_all (testPredicateUnary map ) epl in 
  if test then true else  
  (
   failwith "Input has non unary edb predicates";
  )
  *)
  try (match List.find (testPredicateNotUnary map) epl with Pred(p) -> failwith ("Input has non unary edb predicate: "^p))
  with Not_found -> true                                                                    
)  

let finddiff (rs, ls) = 
(
 let edps = Set.Difference(rs,ls) in 
 edps
)

let printSet s = 
(
	Set.iter (
		fun x -> printf "%a " output_any ( x )
	) s ;
)

let findEDBPredicates (r, l) = 
(
 let rs = Set.of_list(r) in
 let ls = Set.of_list(l) in
 let edps = finddiff(rs,ls) in
 //printSet ls;
 //printSet rs;
 edps 
)

let getPredLists (cList: Clause List) =
(
  let rhslist = ref [] in
  let lhslist = ref [] in 
  List.iter ( fun x -> match x with
			| Edb(a) -> (
			              match a with 
			              |PosUp(BaseUnary(p,q)) -> rhslist := p :: (!rhslist)
			              )
			| Rule(a,b,c) -> ( match a with 
			                  |PosLit(Base(p,q)) -> lhslist := p:: (!lhslist);
			                  List.iter (fun x -> (match x with 
			                                     |PosLit(Base(p,q)) -> rhslist := p::(!rhslist))) b; 		
			                  List.iter (fun x -> (match x with 
			                                     |NegLit(Base(p,q)) -> rhslist := p::(!rhslist))) c;  
			                 ) 
			|  NewRule(a,b) -> List.iter (fun x -> match x with 
		                                  | PosUp(BaseUnary(p,q)) -> rhslist := p::(!rhslist)) a 
		    |  NextRule(a,b,c,d) ->  (
		                             List.iter (fun x -> match x with 
		                                  | PosUp(BaseUnary(p,q)) -> rhslist := p::(!rhslist)) a;
		                             List.iter (fun x -> match x with 
		                                  | NegUp(BaseUnary(p,q)) -> rhslist := p::(!rhslist)) b;
		                             List.iter(fun x -> match x with 
		                                  |PosLit(Base(p,q)) -> rhslist := p::(!rhslist)) c;
		                             List.iter (fun x -> match x with 
		                                  |NegLit(Base(p,q)) -> rhslist := p::(!rhslist)) d; 
		                              )                                                                                                              			                  
			| _ -> ()
			) cList;		
	(!rhslist,!lhslist);	
) 

let findandtestEDBPredicates (pclist: P_Clause List) =
(
 let map = checkPredicateNaming pclist in
 let clist = transformParseOutToProgram (pclist) in 
 let (r,l) = getPredLists clist in
 let edp = findEDBPredicates (r,l) in 
 let test = checkEDBUnary (edp, map) in 
 if test then edp 
    else failwith "Should not print this" 
)